from .local.r2r_local_vector_db import R2RLocalVectorDB
from .pgvector.pgvector_db import PGVectorDB
from .qdrant.qdrant_db import QdrantDB

__all__ = [
    "R2RLocalVectorDB",
    "PGVectorDB",
    "QdrantDB",
]



import json
import logging
import os
from typing import Any

from ..core.abstractions.document import DocumentType
from ..core.pipes.pipe_logging import LoggingConfig
from ..core.providers.embedding_provider import EmbeddingConfig
from ..core.providers.eval_provider import EvalConfig
from ..core.providers.llm_provider import LLMConfig
from ..core.providers.prompt_provider import PromptConfig
from ..core.providers.vector_db_provider import VectorDBConfig

logger = logging.getLogger(__name__)


class R2RConfig:
    REQUIRED_KEYS: dict[str, list] = {
        "app": [
            "max_file_size_in_mb",
        ],
        "embedding": [
            "provider",
            "search_model",
            "search_dimension",
            "batch_size",
            "text_splitter",
        ],
        "eval": [
            "llm",
            "sampling_fraction",
        ],
        "ingestion": ["selected_parsers"],
        "completions": ["provider"],
        "logging": ["provider", "log_table"],
        "prompt": ["provider"],
        "vector_database": ["provider", "collection_name"],
    }
    app: dict[str, Any]
    embedding: EmbeddingConfig
    completions: LLMConfig
    logging: LoggingConfig
    prompt: PromptConfig
    vector_database: VectorDBConfig

    def __init__(self, config_data: dict[str, Any]):
        # Load the default configuration
        default_config = self.load_default_config()

        # Override the default configuration with the passed configuration
        for key in config_data:
            if key in default_config:
                default_config[key].update(config_data[key])
            else:
                default_config[key] = config_data[key]

        # Validate and set the configuration
        for section, keys in R2RConfig.REQUIRED_KEYS.items():
            self._validate_config_section(default_config, section, keys)
            setattr(self, section, default_config[section])

        self.app = self.app  # for type hinting
        self.ingestion = self.ingestion  # for type hinting
        self.ingestion["selected_parsers"] = {
            DocumentType(k): v
            for k, v in self.ingestion["selected_parsers"].items()
        }
        self.embedding = EmbeddingConfig.create(**self.embedding)
        eval_llm = self.eval.pop("llm")
        self.eval = EvalConfig.create(
            **self.eval, llm=LLMConfig.create(**eval_llm)
        )
        self.completions = LLMConfig.create(**self.completions)
        self.logging = LoggingConfig.create(**self.logging)
        self.prompt = PromptConfig.create(**self.prompt)
        self.vector_database = VectorDBConfig.create(**self.vector_database)

    def _validate_config_section(
        self, config_data: dict[str, Any], section: str, keys: list
    ):
        if section not in config_data:
            raise ValueError(f"Missing '{section}' section in config")
        if not all(key in config_data[section] for key in keys):
            raise ValueError(f"Missing required keys in '{section}' config")

    @classmethod
    def from_json(cls, config_path: str = None) -> "R2RConfig":
        if config_path is None:
            # Get the root directory of the project
            file_dir = os.path.dirname(os.path.abspath(__file__))
            config_path = os.path.join(file_dir, "..", "..", "config.json")

        logger.info(f"Loading configuration from {config_path}")

        # Load configuration from JSON file
        with open(config_path) as f:
            config_data = json.load(f)

        return cls(config_data)

    # TODO - How to type 'redis.Redis' without introducing dependency on 'redis' package?
    def save_to_redis(self, redis_client: Any, key: str):
        config_data = {
            section: getattr(self, section)
            for section in R2RConfig.REQUIRED_KEYS.keys()
        }
        redis_client.set(f"R2RConfig:{key}", json.dumps(config_data))

    @classmethod
    def load_from_redis(cls, redis_client: Any, key: str) -> "R2RConfig":
        config_data = redis_client.get(f"R2RConfig:{key}")
        if config_data is None:
            raise ValueError(
                f"Configuration not found in Redis with key '{key}'"
            )
        return cls(json.loads(config_data))

    @classmethod
    def load_default_config(cls) -> dict:
        # Get the root directory of the project
        file_dir = os.path.dirname(os.path.abspath(__file__))
        default_config_path = os.path.join(file_dir, "..", "..", "config.json")
        # Load default configuration from JSON file
        with open(default_config_path) as f:
            return json.load(f)



import argparse
import os

from r2r import (
    R2RApp,
    R2RConfig,
    R2RPipeFactory,
    R2RPipelineFactory,
    R2RProviderFactory,
)

current_file_path = os.path.dirname(__file__)
configs_path = os.path.join(current_file_path, "..", "configs")


def default_app():  # config_name: str = "default", pipe_name: str = "qna"):
    # config_name = os.getenv("CONFIG_OPTION") or config_name
    # pipe_name = os.getenv("PIPELINE_OPTION") or pipe_name

    config = R2RConfig.from_json()

    providers = R2RProviderFactory(config).create_providers()
    pipes = R2RPipeFactory(config, providers).create_pipes()
    pipelines = R2RPipelineFactory(config, pipes).create_pipelines()

    r2r = R2RApp(
        config=config,
        providers=providers,
        pipelines=pipelines,
    )

    return r2r


r2r = default_app()  # args.config)
app = r2r.app

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="R2R Pipe")
    parser.add_argument(
        "--host",
        type=str,
        default="0.0.0.0",
        help="Port to serve deployed pipe on.",
    )
    parser.add_argument(
        "--port",
        type=str,
        default="8000",
        help="Port to serve deployed pipe on.",
    )
    # parser.add_argument(
    #     "--config",
    #     type=str,
    #     default="default",
    #     choices=CONFIG_OPTIONS.keys(),
    #     help="Configuration option for the pipe",
    # )

    args, _ = parser.parse_known_args()

    host = os.getenv("HOST") or args.host
    port = os.getenv("PORT") or args.port

    r2r.serve()



"""A complete demo class for the R2R library."""

import asyncio
import json
import logging
import os
import time
from typing import Optional

import fire
from fastapi.datastructures import UploadFile

from r2r import (
    Document,
    GenerationConfig,
    R2RApp,
    R2RClient,
    R2RConfig,
    R2RPipeFactory,
    R2RPipelineFactory,
    R2RProviderFactory,
    generate_id_from_label,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class R2RDemo:
    """A demo class for the R2R library."""

    DEMO_USER_ID = "063edaf8-3e63-4cb9-a4d6-a855f36376c3"

    def __init__(
        self,
        config_path: Optional[str] = None,
        file_list: Optional[list[str]] = None,
        user_id: str = DEMO_USER_ID,
        base_url: Optional[str] = None,
    ):
        if base_url:
            self.client = R2RClient(base_url)
        else:
            config = R2RConfig.from_json(config_path=config_path)

            providers = R2RProviderFactory(config).create_providers()
            pipes = R2RPipeFactory(config, providers).create_pipes()
            pipelines = R2RPipelineFactory(config, pipes).create_pipelines()
            self.r2r = R2RApp(config, providers, pipelines)

        root_path = os.path.dirname(os.path.abspath(__file__))
        self.user_id = user_id
        self.default_files = file_list or [
            os.path.join(root_path, "data", "screen_shot.png"),
            os.path.join(root_path, "data", "aristotle.txt"),
            os.path.join(root_path, "data", "pg_essay_1.html"),
            os.path.join(root_path, "data", "pg_essay_2.html"),
            os.path.join(root_path, "data", "pg_essay_3.html"),
            os.path.join(root_path, "data", "pg_essay_4.html"),
            os.path.join(root_path, "data", "pg_essay_5.html"),
            os.path.join(root_path, "data", "lyft_2021.pdf"),
            os.path.join(root_path, "data", "uber_2021.pdf"),
        ]

    def ingest_as_documents(self, file_paths: Optional[list[str]] = None):
        file_paths = file_paths or self.default_files
        documents = []
        for file_path in file_paths:
            with open(file_path, "rb") as f:
                data = f.read()

            documents.append(
                Document(
                    id=generate_id_from_label(file_path),
                    data=data,
                    type=file_path.split(".")[-1],
                    metadata={
                        "title": file_path.split(os.path.sep)[-1],
                        "user_id": self.user_id,
                    },
                )
            )

        if hasattr(self, "client"):
            documents_dicts = [doc.dict() for doc in documents]
            t0 = time.time()
            response = self.client.ingest_documents(documents_dicts)
            t1 = time.time()
            print(f"Time taken to ingest files: {t1-t0:.2f} seconds")
            print(response)
        else:
            t0 = time.time()
            self.r2r.ingest_documents(documents)
            t1 = time.time()
            print(f"Time taken to ingest files: {t1-t0:.2f} seconds")

    def ingest_as_files(self, file_paths: Optional[list[str]] = None):
        file_paths = file_paths or self.default_files
        files = [
            UploadFile(
                filename=file_path.split(os.path.sep)[-1],
                file=open(file_path, "rb"),
            )
            for file_path in file_paths
        ]

        # Set file size manually
        for file in files:
            file.file.seek(0, 2)  # Move to the end of the file
            file.size = file.file.tell()  # Get the file size
            file.file.seek(0)  # Move back to the start of the file

        metadatas = [
            {
                "title": file_path.split(os.path.sep)[-1],
                "user_id": self.user_id,
            }
            for file_path in file_paths
        ]

        if hasattr(self, "client"):
            t0 = time.time()
            response = self.client.ingest_files(
                metadatas=metadatas, files=file_paths
            )
            t1 = time.time()
            print(f"Time taken to ingest files: {t1-t0:.2f} seconds")
            print(response)
        else:
            t0 = time.time()
            self.r2r.ingest_files(files=files, metadatas=metadatas)
            t1 = time.time()
            print(f"Time taken to ingest files: {t1-t0:.2f} seconds")

    def search(self, query: str):
        if hasattr(self, "client"):
            t0 = time.time()
            results = self.client.search(
                query, search_filters={"user_id": self.user_id}
            )
            t1 = time.time()
            print(f"Time taken to search: {t1-t0:.2f} seconds")
            for result in results["results"]:
                print(result)
        else:
            t0 = time.time()
            results = self.r2r.search(
                query, search_filters={"user_id": self.user_id}
            )
            t1 = time.time()
            print(f"Time taken to search: {t1-t0:.2f} seconds")
            for result in results["results"]:
                print(result)

    def rag(
        self,
        query: str,
        rag_generation_config: Optional[dict] = None,
        streaming: bool = False,
    ):
        if hasattr(self, "client"):
            if not streaming:
                t0 = time.time()
                response = self.client.rag(
                    query,
                    search_filters={"user_id": self.user_id},
                    rag_generation_config=rag_generation_config,
                    streaming=streaming,
                )
                t1 = time.time()
                print(f"Time taken to get RAG response: {t1-t0:.2f} seconds")
                print(response)
            else:
                t0 = time.time()

                response = self.client.rag(
                    query,
                    search_filters={"user_id": self.user_id},
                    rag_generation_config=rag_generation_config,
                    streaming=streaming,
                )
                collector = ""
                for chunk in response:
                    collector += chunk
                t1 = time.time()
                print(
                    f"\nTime taken to stream RAG response: {t1-t0:.2f} seconds"
                )
        else:
            t0 = time.time()
            rag_generation_config = (
                GenerationConfig(**rag_generation_config, streaming=streaming)
                if rag_generation_config
                else GenerationConfig(
                    **{"stream": streaming, "model": "gpt-3.5-turbo"}
                )
            )
            response = self.r2r.rag(
                query,
                search_filters={"user_id": self.user_id},
                rag_generation_config=rag_generation_config,
            )

            if not streaming:
                t1 = time.time()
                print(f"Time taken to get RAG response: {t1-t0:.2f} seconds")
                print(response)
            else:

                async def _stream_response():
                    async for chunk in response:
                        print(chunk, end="", flush=True)

                asyncio.run(_stream_response())
                t1 = time.time()
                print(
                    f"\nTime taken to stream RAG response: {t1-t0:.2f} seconds"
                )

    def evaluate(
        self,
        query: Optional[str] = None,
        context: Optional[str] = None,
        completion: Optional[str] = None,
        eval_generation_config: Optional[dict] = None,
    ):
        if not query:
            query = "What is the meaning of life?"
        if not context:
            context = """Search Results:
            1. The meaning of life is 42.
            2. The car is red.
            3. The meaning of life is to help others.
            4. The car is blue.
            5. The meaning of life is to learn and grow.
            6. The car is green.
            7. The meaning of life is to make a difference.
            8. The car is yellow.
            9. The meaning of life is to enjoy the journey.
            10. The car is black.
            """
        if not completion:
            completion = "The meaning of life is to help others, learn and grow, and to make a difference."

        if hasattr(self, "client"):
            t0 = time.time()
            response = self.client.evaluate(
                query=query,
                context=context,
                completion=completion,
            )
            t1 = time.time()
            print(f"Time taken to evaluate: {t1-t0:.2f} seconds")
            print(response)
        else:
            t0 = time.time()
            response = self.r2r.evaluate(
                query=query,
                context=context,
                completion=completion,
                eval_generation_config=(
                    GenerationConfig(**eval_generation_config)
                    if eval_generation_config
                    else None
                ),
            )
            t1 = time.time()
            print(f"Time taken to evaluate: {t1-t0:.2f} seconds")
            print(response)

    def delete(
        self,
        key: str = "document_id",
        value: str = "15255e98-e245-5b58-a57f-6c51babf72dd",
    ):
        if hasattr(self, "client"):
            t0 = time.time()
            response = self.client.delete(key, value)
            t1 = time.time()
            print(f"Time taken to delete: {t1-t0:.2f} seconds")
            print(response)
        else:
            t0 = time.time()
            response = self.r2r.delete(key, value)
            t1 = time.time()
            print(f"Time taken to delete: {t1-t0:.2f} seconds")
            print(response)

    def get_user_ids(self):
        if hasattr(self, "client"):
            t0 = time.time()
            response = self.client.get_user_ids()
            t1 = time.time()
            print(f"Time taken to get user IDs: {t1-t0:.2f} seconds")
            print(response)
        else:
            t0 = time.time()
            response = self.r2r.get_user_ids()
            t1 = time.time()
            print(f"Time taken to get user IDs: {t1-t0:.2f} seconds")
            print(response)

    def get_user_document_data(self):
        if hasattr(self, "client"):
            t0 = time.time()
            response = self.client.get_user_document_data(self.user_id)
            t1 = time.time()
            print(f"Time taken to get user document data: {t1-t0:.2f} seconds")
            print(response)
        else:
            t0 = time.time()
            response = self.r2r.get_user_document_data(self.user_id)
            t1 = time.time()
            print(f"Time taken to get user document data: {t1-t0:.2f} seconds")
            print(response)

    def get_logs(self, pipeline_type: Optional[str] = None):
        if hasattr(self, "client"):
            t0 = time.time()
            response = self.client.get_logs(pipeline_type)
            t1 = time.time()
            print(f"Time taken to get logs: {t1-t0:.2f} seconds")
            print(response)
        else:
            t0 = time.time()
            response = self.r2r.get_logs(pipeline_type)
            t1 = time.time()
            print(f"Time taken to get logs: {t1-t0:.2f} seconds")
            print(response)

    def get_open_api_endpoint(self):
        if hasattr(self, "client"):
            print(
                json.dumps(
                    self.client.get_open_api_endpoint()["results"], indent=2
                )
            )
        else:
            print(
                json.dumps(
                    self.r2r.get_open_api_endpoint()["results"], indent=2
                )
            )

    def serve(self, host: str = "0.0.0.0", port: int = 8000):
        self.r2r.serve(host, port)


if __name__ == "__main__":
    fire.Fire(R2RDemo)



import asyncio
import uuid
from typing import TYPE_CHECKING, Any, AsyncGenerator, Iterable

if TYPE_CHECKING:
    from ..pipeline.base_pipeline import Pipeline


def generate_run_id() -> uuid.UUID:
    return uuid.uuid4()


def generate_id_from_label(label: str) -> uuid.UUID:
    return uuid.uuid5(uuid.NAMESPACE_DNS, label)


async def to_async_generator(
    iterable: Iterable[Any],
) -> AsyncGenerator[Any, None]:
    for item in iterable:
        yield item


def run_pipeline(pipeline: "Pipeline", input: Any, *args, **kwargs):
    if not isinstance(input, AsyncGenerator) and not isinstance(input, list):
        input = to_async_generator([input])
    elif not isinstance(input, AsyncGenerator):
        input = to_async_generator(input)

    async def _run_pipeline(input, *args, **kwargs):
        return await pipeline.run(input, *args, **kwargs)

    return asyncio.run(_run_pipeline(input, *args, **kwargs))



import json
import logging
import os
import sqlite3
from typing import Optional, Union

from r2r.core import (
    SearchResult,
    VectorDBConfig,
    VectorDBProvider,
    VectorEntry,
)

logger = logging.getLogger(__name__)


class R2RLocalVectorDB(VectorDBProvider):
    def __init__(self, config: VectorDBConfig) -> None:
        super().__init__(config)
        if config.provider != "local":
            raise ValueError(
                "R2RLocalVectorDB must be initialized with provider `local`."
            )

    def _get_conn(self):
        conn = sqlite3.connect(
            self.config.extra_fields.get("db_path")
            or os.getenv("LOCAL_DB_PATH", "local.sqlite")
        )
        return conn

    def _get_cursor(self, conn):
        return conn.cursor()

    def initialize_collection(self, dimension: int) -> None:
        conn = self._get_conn()
        cursor = self._get_cursor(conn)
        cursor.execute(
            f"""
            CREATE TABLE IF NOT EXISTS "{self.config.collection_name}" (
                id TEXT PRIMARY KEY,
                vector TEXT,
                metadata TEXT
            )
        """
        )
        conn.commit()
        conn.close()

    def create_index(self, index_type, column_name, index_options):
        raise NotImplementedError(
            "R2RLocalVectorDB does not support creating indexes."
        )

    def copy(self, entry: VectorEntry, commit=True) -> None:
        if self.config.collection_name is None:
            raise ValueError(
                "Collection name is not set. Please call `initialize_collection` first."
            )

        conn = self._get_conn()
        cursor = self._get_cursor(conn)
        serializeable_entry = entry.to_serializable()
        cursor.execute(
            f"""
                INSERT OR IGNORE INTO "{self.config.collection_name}" (id, vector, metadata)
                VALUES (?, ?, ?)
            """,
            (
                serializeable_entry["id"],
                str(serializeable_entry["vector"]),
                json.dumps(serializeable_entry["metadata"]),
            ),
        )
        if commit:
            conn.commit()
        conn.close()

    def upsert(self, entry: VectorEntry, commit=True) -> None:
        if self.config.collection_name is None:
            raise ValueError(
                "Collection name is not set. Please call `initialize_collection` first."
            )

        conn = self._get_conn()
        cursor = self._get_cursor(conn)
        serializeable_entry = entry.to_serializable()
        cursor.execute(
            f"""
                INSERT OR REPLACE INTO "{self.config.collection_name}" (id, vector, metadata)
                VALUES (?, ?, ?)
            """,
            (
                serializeable_entry["id"],
                str(serializeable_entry["vector"]),
                json.dumps(serializeable_entry["metadata"]),
            ),
        )
        if commit:
            conn.commit()
        conn.close()

    def _cosine_similarity(
        self, vec1: list[float], vec2: list[float]
    ) -> float:
        dot_product = sum(a * b for a, b in zip(vec1, vec2))
        norm_a = sum(a * a for a in vec1) ** 0.5
        norm_b = sum(b * b for b in vec2) ** 0.5
        return dot_product / (norm_a * norm_b)

    def search(
        self,
        query_vector: list[float],
        filters: dict[str, Union[bool, int, str]] = {},
        limit: int = 10,
        *args,
        **kwargs,
    ) -> list[SearchResult]:
        if self.config.collection_name is None:
            raise ValueError(
                "Collection name is not set. Please call `initialize_collection` first."
            )
        conn = self._get_conn()
        cursor = self._get_cursor(conn)
        cursor.execute(f'SELECT * FROM "{self.config.collection_name}"')
        results = []
        for id, vector, metadata in cursor.fetchall():
            vector = json.loads(vector)
            json_metadata = json.loads(metadata)
            if all(json_metadata.get(k) == v for k, v in filters.items()):
                # Local cosine similarity calculation
                score = self._cosine_similarity(query_vector, vector)
                results.append(
                    SearchResult(id=id, score=score, metadata=json_metadata)
                )
        results.sort(key=lambda x: x.score, reverse=True)
        conn.close()
        return results[:limit]

    def delete_by_metadata(
        self, metadata_field: str, metadata_value: Union[bool, int, str]
    ) -> None:
        if self.config.collection_name is None:
            raise ValueError(
                "Collection name is not set. Please call `initialize_collection` first."
            )

        conn = self._get_conn()
        cursor = self._get_cursor(conn)
        cursor.execute(f'SELECT * FROM "{self.config.collection_name}"')
        for id, vector, metadata in cursor.fetchall():
            metadata_json = json.loads(metadata)
            if metadata_json.get(metadata_field) == metadata_value:
                cursor.execute(
                    f'DELETE FROM "{self.config.collection_name}" WHERE id = ?',
                    (id,),
                )
        conn.commit()
        conn.close()

    def get_metadatas(
        self,
        metadata_fields: list[str],
        filter_field: Optional[str] = None,
        filter_value: Optional[str] = None,
    ) -> list[str]:
        if self.config.collection_name is None:
            raise ValueError(
                "Collection name is not set. Please call `initialize_collection` first."
            )
        conn = self._get_conn()
        cursor = self._get_cursor(conn)
        cursor.execute(f'SELECT metadata FROM "{self.config.collection_name}"')
        results = set([])
        for (metadata,) in cursor.fetchall():
            metatada_json = json.loads(metadata)
            if (
                filter_field is None
                or metatada_json.get(filter_field) == filter_value
            ):
                results.add(
                    json.dumps(
                        {
                            k: metatada_json.get(k, None)
                            for k in metadata_fields
                        }
                    )
                )
        conn.close()
        return [json.loads(r) for r in results]



"""A module for creating LlamaCpp model abstractions."""

import logging
import os
from typing import Union

from r2r.core import (
    GenerationConfig,
    LLMChatCompletion,
    LLMChatCompletionChunk,
    LLMConfig,
    LLMProvider,
)

logger = logging.getLogger(__name__)


class LlamaCppConfig(LLMConfig):
    """Configuration for LlamaCpp models."""

    # Base
    provider: str = "llama-cpp"
    model: str = ""
    path_to_model: str = ""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        if not self.path_to_model or self.path_to_model == "":
            self.path_to_model = os.path.join(
                os.path.expanduser("~"), ".cache", "models"
            )
        if not self.model or self.model == "":
            self.model = "tinyllama-1.1b-chat-v1.0.Q2_K.gguf"


class LlamaCPP(LLMProvider):
    """A concrete class for creating LlamaCpp models."""

    def __init__(
        self,
        config: LlamaCppConfig,
        *args,
        **kwargs,
    ) -> None:
        if not isinstance(config, LlamaCppConfig):
            raise ValueError(
                "The provided config must be an instance of LlamaCppConfig."
            )
        self.config: LlamaCppConfig = config

        try:
            from llama_cpp import Llama
        except ImportError:
            raise ImportError(
                "Error, `llama-cpp-python` is required to run a LlamaCPP. Please install it using `pip install llama-cpp-python`."
            )

        super().__init__(config)
        path = os.path.join(self.config.path_to_model, self.config.model)
        self.client = Llama(path, n_ctx=2048)

    def get_completion(
        self,
        messages: list[dict],
        generation_config: GenerationConfig,
        **kwargs,
    ) -> LLMChatCompletion:
        if generation_config.stream:
            raise ValueError(
                "Stream must be set to False to use the `get_completion` method."
            )
        return self._get_completion(messages, generation_config, **kwargs)

    def get_completion_stream(
        self,
        messages: list[dict],
        generation_config: GenerationConfig,
        **kwargs,
    ) -> LLMChatCompletionChunk:
        if not generation_config.stream:
            raise ValueError(
                "Stream must be set to True to use the `get_completion_stream` method."
            )
        return self._get_completion(messages, generation_config, **kwargs)

    def _get_completion(
        self,
        messages: list[dict],
        generation_config: GenerationConfig,
        **kwargs,
    ) -> Union[LLMChatCompletion, LLMChatCompletionChunk]:
        """Get a completion from the LlamaCpp model based on the provided messages."""

        args = self._get_base_args(generation_config)

        prompt = "\n".join([msg["content"] for msg in messages])

        response = self.client(prompt, **args)

        if not generation_config.stream:
            return response
        else:
            return LLMChatCompletionChunk(
                choices=[
                    {
                        "message": {
                            "role": "assistant",
                            "content": str(response),
                        },
                        "index": 0,
                        "finish_reason": None,
                    }
                ],
                **kwargs,
            )

    def _get_base_args(
        self,
        generation_config: GenerationConfig,
    ) -> dict:
        """Get the base arguments for the LlamaCpp model."""
        args = {
            "temperature": generation_config.temperature,
            "top_p": generation_config.top_p,
            "max_tokens": generation_config.max_tokens_to_sample,
        }
        return args

    def extract_content(self, response: LLMChatCompletion) -> str:
        return response.choices[0].message.content



from .litellm.base_litellm import LiteLLM
from .llama_cpp.base_llama_cpp import LlamaCPP, LlamaCppConfig
from .openai.base_openai import OpenAILLM

__all__ = [
    "LiteLLM",
    "LlamaCPP",
    "LlamaCppConfig",
    "OpenAILLM",
]



"""Module for the R2RClient class."""

import asyncio
import json
from typing import AsyncGenerator, Generator, Optional, Union

import httpx
import nest_asyncio
import requests

nest_asyncio.apply()


class R2RClient:
    def __init__(self, base_url: str):
        self.base_url = base_url

    def ingest_documents(self, documents: list[dict]) -> dict:
        url = f"{self.base_url}/ingest_documents/"
        data = {"documents": documents}
        response = requests.post(url, json=data)
        response.raise_for_status()
        return response.json()

    def ingest_files(
        self,
        metadatas: list[dict],
        files: list[str],
        ids: Optional[list[str]] = None,
    ) -> dict:
        url = f"{self.base_url}/ingest_files/"
        files_to_upload = [
            ("files", (file, open(file, "rb"), "application/octet-stream"))
            for file in files
        ]
        data = {
            "metadatas": None if not metadatas else json.dumps(metadatas),
            "ids": None if not ids else json.dumps(ids),
        }
        response = requests.post(url, files=files_to_upload, data=data)
        response.raise_for_status()
        return response.json()

    def search(
        self, query: str, search_filters: dict = {}, search_limit: int = 10
    ) -> dict:
        url = f"{self.base_url}/search/"
        data = {
            "query": query,
            "search_filters": json.dumps(search_filters),
            "search_limit": search_limit,
        }
        response = requests.post(url, json=data)
        response.raise_for_status()
        return response.json()

    def rag(
        self,
        message: str,
        search_filters: Optional[dict] = None,
        search_limit: int = 10,
        rag_generation_config: Optional[dict] = None,
        streaming: bool = False,
    ) -> Union[dict, Generator[str, None, None]]:
        if streaming:
            return self._stream_rag_sync(
                message=message,
                search_filters=search_filters,
                search_limit=search_limit,
                rag_generation_config=rag_generation_config,
            )
        else:
            url = f"{self.base_url}/rag/"
            data = {
                "message": message,
                "search_filters": json.dumps(search_filters or {}),
                "search_limit": search_limit,
                "streaming": streaming,
            }
            if rag_generation_config:
                data["rag_generation_config"] = json.dumps(
                    rag_generation_config
                )
            response = requests.post(url, json=data)
            response.raise_for_status()
            return response.json()

    async def _stream_rag(
        self,
        message: str,
        search_filters: Optional[dict] = None,
        search_limit: int = 10,
        rag_generation_config: Optional[dict] = None,
    ) -> Generator[str, None, None]:
        url = f"{self.base_url}/rag/"
        data = {
            "message": message,
            "search_filters": json.dumps(search_filters or {}),
            "search_limit": search_limit,
            "streaming": True,
        }
        if rag_generation_config:
            data["rag_generation_config"] = json.dumps(rag_generation_config)
        async with httpx.AsyncClient() as client:
            async with client.stream("POST", url, json=data) as response:
                response.raise_for_status()
                async for chunk in response.aiter_text():
                    yield chunk

    def _stream_rag_sync(
        self,
        message: str,
        search_filters: Optional[dict] = None,
        search_limit: int = 10,
        rag_generation_config: Optional[dict] = None,
    ) -> Generator[str, None, None]:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        async_gen = self._stream_rag(
            message=message,
            search_filters=search_filters,
            search_limit=search_limit,
            rag_generation_config=rag_generation_config,
        )
        for chunk in loop.run_until_complete(
            self._iterate_async_gen(async_gen)
        ):
            yield chunk

    async def _iterate_async_gen(
        self, async_gen: AsyncGenerator[str, None]
    ) -> list[str]:
        chunks = []
        async for chunk in async_gen:
            chunks.append(chunk)
        return chunks

    def delete(self, key: str, value: str) -> dict:
        url = f"{self.base_url}/delete/"
        data = {"key": key, "value": value}
        response = requests.request("DELETE", url, json=data)
        response.raise_for_status()
        return response.json()

    def get_user_ids(self) -> dict:
        url = f"{self.base_url}/get_user_ids/"
        response = requests.get(url)
        response.raise_for_status()
        return response.json()

    def get_user_document_data(self, user_id: str) -> dict:
        url = f"{self.base_url}/get_user_document_data/"
        data = {"user_id": user_id}
        response = requests.post(url, json=data)
        response.raise_for_status()
        response_json = response.json()
        return response_json

    def get_logs(
        self, pipeline_type: Optional[str] = None, filter: Optional[str] = None
    ) -> dict:
        url = f"{self.base_url}/get_logs/"
        data = {"pipeline_type": pipeline_type, "filter": filter}
        response = requests.post(url, json=data)
        response.raise_for_status()
        return response.json()



import asyncio
import uuid
from typing import Any, AsyncGenerator, Optional

from .base_pipe import AsyncPipe, AsyncState, PipeType
from .pipe_logging import PipeLoggingConnectionSingleton


class LoggableAsyncPipe(AsyncPipe):
    """An asynchronous pipe for processing data with logging capabilities."""

    class PipeConfig(AsyncPipe.PipeConfig):
        """Configuration for a loggable pipe."""

        name: str = "default_loggable_pipe"
        max_log_queue_size: int = 100

    def __init__(
        self,
        pipe_logger: Optional[PipeLoggingConnectionSingleton] = None,
        type: PipeType = PipeType.OTHER,
        config: Optional[AsyncPipe.PipeConfig] = None,
        *args,
        **kwargs
    ):
        if not pipe_logger:
            pipe_logger = PipeLoggingConnectionSingleton()
        self.pipe_logger = pipe_logger
        self.log_queue = asyncio.Queue()
        self.log_worker_task = None

        super().__init__(type=type, config=config, *args, **kwargs)

    async def log_worker(self):
        while True:
            log_data = await self.log_queue.get()
            pipe_run_id, key, value = log_data
            await self.pipe_logger.log(pipe_run_id, key, value)
            self.log_queue.task_done()

    async def enqueue_log(self, pipe_run_id: str, key: str, value: str):
        if self.log_queue.qsize() < self.config.max_log_queue_size:
            await self.log_queue.put((pipe_run_id, key, value))

    async def run(
        self,
        input: AsyncPipe.Input,
        state: AsyncState,
        run_id: Optional[uuid.UUID] = None,
        *args: Any,
        **kwargs: Any
    ) -> AsyncGenerator[Any, None]:
        """Run the pipe with logging capabilities."""

        async def wrapped_run() -> AsyncGenerator[Any, None]:
            await self._initiate_run(run_id)
            self.log_worker_task = asyncio.create_task(self.log_worker())
            try:
                async for result in self._run_logic(
                    input, state, *args, **kwargs
                ):
                    yield result
            finally:
                await self.log_queue.join()
                self.log_worker_task.cancel()

        return wrapped_run()



import random
from abc import abstractmethod
from typing import Optional, Union

from .base_provider import Provider, ProviderConfig
from .llm_provider import LLMConfig


class EvalConfig(ProviderConfig):
    """A base eval config class"""

    sampling_fraction: float = 0.0
    llm: LLMConfig

    def validate(self) -> None:
        if self.provider not in self.supported_providers:
            raise ValueError(f"Provider {self.provider} not supported.")

    @property
    def supported_providers(self) -> list[str]:
        return ["local"]


class EvalProvider(Provider):
    """An abstract class to provide a common interface for evaluation providers."""

    def __init__(self, config: EvalConfig):
        if not isinstance(config, EvalConfig):
            raise ValueError(
                "EvalProvider must be initialized with a `EvalConfig`."
            )

        super().__init__(config)

    @abstractmethod
    def _evaluate(
        self, query: str, context: str, completion: str
    ) -> dict[str, dict[str, Union[str, float]]]:
        pass

    def evaluate(
        self, query: str, context: str, completion: str
    ) -> Optional[dict]:
        if random.random() < self.config.sampling_fraction:
            return self._evaluate(query, context, completion)
        return None



"""Abstractions for documents and their extractions."""

import uuid
from enum import Enum
from typing import Union

from pydantic import BaseModel

DataType = Union[str, bytes]


class DocumentType(Enum):
    """Types of documents that can be stored."""

    CSV = "csv"
    DOCX = "docx"
    HTML = "html"
    JSON = "json"
    MD = "md"
    PDF = "pdf"
    PPTX = "pptx"
    TXT = "txt"
    XLSX = "xlsx"
    GIF = "gif"
    PNG = "png"
    JPG = "jpg"
    JPEG = "jpeg"
    SVG = "svg"
    MP3 = "mp3"
    MP4 = "mp4"


class Document(BaseModel):
    """A document that has been stored in the system."""

    id: uuid.UUID
    type: DocumentType
    data: DataType
    metadata: dict


class ExtractionType(Enum):
    """Types of extractions that can be performed."""

    TXT = "txt"
    IMG = "img"
    MOV = "mov"


class Extraction(BaseModel):
    """An extraction from a document."""

    id: uuid.UUID
    type: ExtractionType = ExtractionType.TXT
    data: DataType
    metadata: dict
    document_id: uuid.UUID


class FragmentType(Enum):
    """A type of fragment that can be extracted from a document."""

    TEXT = "text"
    IMAGE = "image"


class Fragment(BaseModel):
    """A fragment extracted from a document."""

    id: uuid.UUID
    type: FragmentType
    data: DataType
    metadata: dict
    document_id: uuid.UUID
    extraction_id: uuid.UUID



import logging
import os

from openai import AsyncOpenAI, OpenAI

from r2r.core import EmbeddingConfig, EmbeddingProvider, SearchResult

logger = logging.getLogger(__name__)


class OpenAIEmbeddingProvider(EmbeddingProvider):
    MODEL_TO_TOKENIZER = {
        "text-embedding-ada-002": "cl100k_base",
        "text-embedding-3-small": "cl100k_base",
        "text-embedding-3-large": "cl100k_base",
    }
    MODEL_TO_DIMENSIONS = {
        "text-embedding-ada-002": [1536],
        "text-embedding-3-small": [512, 1536],
        "text-embedding-3-large": [256, 1024, 3072],
    }

    def __init__(self, config: EmbeddingConfig):
        logger.info(
            "Initializing `OpenAIEmbeddingProvider` to provide embeddings."
        )
        super().__init__(config)
        provider = config.provider
        if not provider:
            raise ValueError(
                "Must set provider in order to initialize SentenceTransformerEmbeddingProvider."
            )

        if provider != "openai":
            raise ValueError(
                "OpenAIEmbeddingProvider must be initialized with provider `openai`."
            )
        if not os.getenv("OPENAI_API_KEY"):
            raise ValueError(
                "Must set OPENAI_API_KEY in order to initialize OpenAIEmbeddingProvider."
            )
        self.client = OpenAI()
        self.async_client = AsyncOpenAI()

        if config.rerank_model:
            raise ValueError(
                "OpenAIEmbeddingProvider does not support separate reranking."
            )
        self.search_model = config.search_model
        self.search_dimension = config.search_dimension

        if self.search_model not in OpenAIEmbeddingProvider.MODEL_TO_TOKENIZER:
            raise ValueError(
                f"OpenAI embedding model {self.search_model} not supported."
            )
        if (
            self.search_dimension
            and self.search_dimension
            not in OpenAIEmbeddingProvider.MODEL_TO_DIMENSIONS[
                self.search_model
            ]
        ):
            raise ValueError(
                f"Dimensions {self.dimension} for {self.search_model} are not supported"
            )

        if not self.search_model or not self.search_dimension:
            raise ValueError(
                "Must set search_model and search_dimension in order to initialize OpenAIEmbeddingProvider."
            )

        if config.rerank_model:
            raise ValueError(
                "OpenAIEmbeddingProvider does not support separate reranking."
            )

    def get_embedding(
        self,
        text: str,
        stage: EmbeddingProvider.PipeStage = EmbeddingProvider.PipeStage.SEARCH,
    ) -> list[float]:
        if stage != EmbeddingProvider.PipeStage.SEARCH:
            raise ValueError(
                "OpenAIEmbeddingProvider only supports search stage."
            )

        return (
            self.client.embeddings.create(
                input=[text],
                model=self.search_model,
                dimensions=self.search_dimension
                or OpenAIEmbeddingProvider.MODEL_TO_DIMENSIONS[
                    self.search_model
                ][-1],
            )
            .data[0]
            .embedding
        )

    async def async_get_embedding(
        self,
        text: str,
        stage: EmbeddingProvider.PipeStage = EmbeddingProvider.PipeStage.SEARCH,
    ) -> list[float]:
        if stage != EmbeddingProvider.PipeStage.SEARCH:
            raise ValueError(
                "OpenAIEmbeddingProvider only supports search stage."
            )

        response = await self.async_client.embeddings.create(
            input=[text],
            model=self.search_model,
            dimensions=self.search_dimension
            or OpenAIEmbeddingProvider.MODEL_TO_DIMENSIONS[self.search_model][
                -1
            ],
        )
        return response.data[0].embedding

    def get_embeddings(
        self,
        texts: list[str],
        stage: EmbeddingProvider.PipeStage = EmbeddingProvider.PipeStage.SEARCH,
    ) -> list[list[float]]:
        if stage != EmbeddingProvider.PipeStage.SEARCH:
            raise ValueError(
                "OpenAIEmbeddingProvider only supports search stage."
            )

        return [
            ele.embedding
            for ele in self.client.embeddings.create(
                input=texts,
                model=self.search_model,
                dimensions=self.search_dimension
                or OpenAIEmbeddingProvider.MODEL_TO_DIMENSIONS[
                    self.search_model
                ][-1],
            ).data
        ]

    async def async_get_embeddings(
        self,
        texts: list[str],
        stage: EmbeddingProvider.PipeStage = EmbeddingProvider.PipeStage.SEARCH,
    ) -> list[list[float]]:
        if stage != EmbeddingProvider.PipeStage.SEARCH:
            raise ValueError(
                "OpenAIEmbeddingProvider only supports search stage."
            )

        response = await self.async_client.embeddings.create(
            input=texts,
            model=self.search_model,
            dimensions=self.search_dimension
            or OpenAIEmbeddingProvider.MODEL_TO_DIMENSIONS[self.search_model][
                -1
            ],
        )
        return [ele.embedding for ele in response.data]

    def rerank(
        self,
        transformed_message: str,
        texts: list[SearchResult],
        stage: EmbeddingProvider.PipeStage = EmbeddingProvider.PipeStage.RERANK,
        limit: int = 10,
    ):
        return texts[:limit]

    def tokenize_string(self, text: str, model: str) -> list[int]:
        try:
            import tiktoken
        except ImportError:
            raise ValueError(
                "Must download tiktoken library to run `tokenize_string`."
            )
        # tiktoken encoding -
        # cl100k_base -	gpt-4, gpt-3.5-turbo, text-embedding-ada-002, text-embedding-3-small, text-embedding-3-large
        if model not in OpenAIEmbeddingProvider.MODEL_TO_TOKENIZER:
            raise ValueError(f"OpenAI embedding model {model} not supported.")
        encoding = tiktoken.get_encoding(
            OpenAIEmbeddingProvider.MODEL_TO_TOKENIZER[model]
        )
        return encoding.encode(text)



import logging
import os
from typing import Optional, Union

from r2r.core import (
    SearchResult,
    VectorDBConfig,
    VectorDBProvider,
    VectorEntry,
)
from r2r.vecs.client import Client
from r2r.vecs.collection import Collection

logger = logging.getLogger(__name__)


class PGVectorDB(VectorDBProvider):
    def __init__(self, config: VectorDBConfig) -> None:
        super().__init__(config)
        try:
            import r2r.vecs
        except ImportError:
            raise ValueError(
                f"Error, PGVectorDB requires the vecs library. Please run `poetry add vecs`."
            )
        user = os.getenv("POSTGRES_USER")
        password = os.getenv("POSTGRES_PASSWORD")
        host = os.getenv("POSTGRES_HOST")
        port = os.getenv("POSTGRES_PORT")
        db_name = os.getenv("POSTGRES_DBNAME")
        if not all([user, password, host, port, db_name]):
            raise ValueError(
                "Error, please set the POSTGRES_USER, POSTGRES_PASSWORD, POSTGRES_HOST, POSTGRES_PORT, and POSTGRES_DBNAME environment variables."
            )
        try:
            DB_CONNECTION = (
                f"postgresql://{user}:{password}@{host}:{port}/{db_name}"
            )
            self.vx: Client = r2r.vecs.create_client(DB_CONNECTION)
        except Exception as e:
            raise ValueError(
                f"Error {e} occurred while attempting to connect to the pgvector provider with {DB_CONNECTION}."
            )
        self.collection: Optional[Collection] = None
        self.config: VectorDBConfig = config

    def initialize_collection(self, dimension: int) -> None:
        self.collection = self.vx.get_or_create_collection(
            name=self.config.collection_name, dimension=dimension
        )

    def copy(self, entry: VectorEntry, commit=True) -> None:
        if self.collection is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `copy`."
            )

        serializeable_entry = entry.to_serializable()

        self.collection.copy(
            records=[
                (
                    serializeable_entry["id"],
                    serializeable_entry["vector"],
                    serializeable_entry["metadata"],
                )
            ]
        )

    def copy_entries(
        self, entries: list[VectorEntry], commit: bool = True
    ) -> None:
        if self.collection is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `copy_entries`."
            )

        self.collection.copy(
            records=[
                (
                    str(entry.id),
                    entry.vector.data,
                    entry.to_serializable()["metadata"],
                )
                for entry in entries
            ]
        )

    def upsert(self, entry: VectorEntry, commit=True) -> None:
        if self.collection is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `upsert`."
            )

        self.collection.upsert(
            records=[
                (
                    str(entry.id),
                    entry.vector.data,
                    entry.to_serializable()["metadata"],
                )
            ]
        )

    def upsert_entries(
        self, entries: list[VectorEntry], commit: bool = True
    ) -> None:
        if self.collection is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `upsert_entries`."
            )

        self.collection.upsert(
            records=[
                (
                    str(entry.id),
                    entry.vector.data,
                    entry.to_serializable()["metadata"],
                )
                for entry in entries
            ]
        )

    def search(
        self,
        query_vector: list[float],
        filters: dict[str, Union[bool, int, str]] = {},
        limit: int = 10,
        *args,
        **kwargs,
    ) -> list[SearchResult]:
        if self.collection is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `search`."
            )
        measure = kwargs.get("measure", "cosine_distance")
        mapped_filters = {
            key: {"$eq": value} for key, value in filters.items()
        }

        return [
            SearchResult(id=ele[0], score=float(1 - ele[1]), metadata=ele[2])  # type: ignore
            for ele in self.collection.query(
                data=query_vector,
                limit=limit,
                filters=mapped_filters,
                measure=measure,
                include_value=True,
                include_metadata=True,
            )
        ]

    def create_index(self, index_type, column_name, index_options):
        pass

    def delete_by_metadata(
        self, metadata_field: str, metadata_value: Union[bool, int, str]
    ) -> None:
        if self.collection is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `delete_by_metadata`."
            )
        self.collection.delete(
            filters={metadata_field: {"$eq": metadata_value}}
        )

    def get_metadatas(
        self,
        metadata_fields: list[str],
        filter_field: Optional[str] = None,
        filter_value: Optional[Union[bool, int, str]] = None,
    ) -> list[dict]:
        if self.collection is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `get_metadatas`."
            )

        results = {tuple(metadata_fields): {}}
        for field in metadata_fields:
            unique_values = self.collection.get_unique_metadata_values(
                field=field,
                filter_field=filter_field,
                filter_value=filter_value,
            )
            for value in unique_values:
                if value not in results:
                    results[value] = {}
                results[value][field] = value

        return [
            results[key] for key in results if key != tuple(metadata_fields)
        ]



from .abstractions.document import (
    DataType,
    Document,
    DocumentType,
    Extraction,
    ExtractionType,
    Fragment,
    FragmentType,
)
from .abstractions.llm import LLMChatCompletion, LLMChatCompletionChunk
from .abstractions.prompt import Prompt
from .abstractions.search import SearchRequest, SearchResult
from .abstractions.vector import Vector, VectorEntry, VectorType
from .parsers import (
    AsyncParser,
    AudioParser,
    CSVParser,
    DOCXParser,
    HTMLParser,
    ImageParser,
    JSONParser,
    MarkdownParser,
    MovieParser,
    PDFParser,
    PPTParser,
    TextParser,
    XLSXParser,
)
from .pipeline.base_pipeline import (
    EvalPipeline,
    IngestionPipeline,
    Pipeline,
    RAGPipeline,
    SearchPipeline,
)
from .pipes.base_pipe import AsyncPipe, AsyncState, PipeRunInfo, PipeType
from .pipes.loggable_pipe import LoggableAsyncPipe
from .pipes.pipe_logging import (
    LocalPipeLoggingProvider,
    LoggingConfig,
    PipeLoggingConnectionSingleton,
    PostgresLoggingConfig,
    PostgresPipeLoggingProvider,
    RedisLoggingConfig,
    RedisPipeLoggingProvider,
)
from .providers.embedding_provider import EmbeddingConfig, EmbeddingProvider
from .providers.eval_provider import EvalConfig, EvalProvider
from .providers.llm_provider import GenerationConfig, LLMConfig, LLMProvider
from .providers.prompt_provider import PromptConfig, PromptProvider
from .providers.vector_db_provider import VectorDBConfig, VectorDBProvider
from .utils import (
    RecursiveCharacterTextSplitter,
    TextSplitter,
    generate_id_from_label,
    generate_run_id,
    run_pipeline,
    to_async_generator,
)

__all__ = [
    # Logging
    "LoggingConfig",
    "LocalPipeLoggingProvider",
    "PostgresLoggingConfig",
    "PostgresPipeLoggingProvider",
    "RedisLoggingConfig",
    "RedisPipeLoggingProvider",
    "PipeLoggingConnectionSingleton",
    # Abstractions
    "VectorEntry",
    "VectorType",
    "Vector",
    "SearchRequest",
    "SearchResult",
    "AsyncPipe",
    "PipeRunInfo",
    "PipeType",
    "AsyncState",
    "LoggableAsyncPipe",
    "Prompt",
    "DataType",
    "DocumentType",
    "Document",
    "Extraction",
    "ExtractionType",
    "Fragment",
    "FragmentType",
    # Parsers
    "AudioParser",
    "AsyncParser",
    "CSVParser",
    "DOCXParser",
    "HTMLParser",
    "ImageParser",
    "JSONParser",
    "MarkdownParser",
    "MovieParser",
    "PDFParser",
    "PPTParser",
    "TextParser",
    "XLSXParser",
    # Pipelines
    "Pipeline",
    "EvalPipeline",
    "IngestionPipeline",
    "RAGPipeline",
    "SearchPipeline",
    # Providers
    "EmbeddingConfig",
    "EmbeddingProvider",
    "EvalConfig",
    "EvalProvider",
    "PromptConfig",
    "PromptProvider",
    "GenerationConfig",
    "LLMChatCompletion",
    "LLMChatCompletionChunk",
    "LLMConfig",
    "LLMProvider",
    "VectorDBConfig",
    "VectorDBProvider",
    # Other
    "TextSplitter",
    "RecursiveCharacterTextSplitter",
    "to_async_generator",
    "run_pipeline",
    "generate_run_id",
    "generate_id_from_label",
]



import asyncio
import json
import logging
import uuid
from typing import Optional, Union

from fastapi import Body, FastAPI, File, Form, HTTPException, UploadFile
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import StreamingResponse

from r2r.core import (
    Document,
    DocumentType,
    GenerationConfig,
    PipeLoggingConnectionSingleton,
    generate_id_from_label,
    generate_run_id,
    to_async_generator,
)
from r2r.pipes import R2REvalPipe

from .r2r_abstractions import R2RPipelines, R2RProviders
from .r2r_config import R2RConfig

MB_CONVERSION_FACTOR = 1024 * 1024

from pydantic import BaseModel

logger = logging.getLogger(__name__)


def syncable(func):
    """Decorator to mark methods for synchronous wrapper creation."""
    func._syncable = True
    return func


class AsyncSyncMeta(type):
    """Metaclass to create synchronous wrappers for async methods."""

    def __new__(cls, name, bases, dct):
        new_cls = super().__new__(cls, name, bases, dct)
        for attr_name, attr_value in dct.items():
            if asyncio.iscoroutinefunction(attr_value) and getattr(
                attr_value, "_syncable", False
            ):
                sync_method_name = attr_name[
                    1:
                ]  # Remove leading 'a' for sync method
                async_method = attr_value

                def make_sync_method(async_method):
                    def sync_wrapper(self, *args, **kwargs):
                        try:
                            import nest_asyncio

                            nest_asyncio.apply()
                            result = asyncio.run(
                                async_method(self, *args, **kwargs)
                            )
                            return result
                        except Exception as e:
                            logger.error(
                                f"Error in sync method {sync_method_name}: {str(e)}"
                            )
                            raise

                    return sync_wrapper

                setattr(
                    new_cls, sync_method_name, make_sync_method(async_method)
                )
        return new_cls


class R2RApp(metaclass=AsyncSyncMeta):
    """Main class for the R2R application.

    This class is responsible for setting up the FastAPI application and
    defining the routes for the various endpoints. It also contains the
    synchronous wrappers for the asynchronous methods defined in the class.

    Endpoints are provided to:
    - Ingest documents
    - Ingest files
    - Search
    - Retrieve and generate completions
    - Delete entries
    - Retrieve user IDs
    - Retrieve user document data
    - Retrieve logs
    """

    def __init__(
        self,
        config: R2RConfig,
        providers: R2RProviders,
        pipelines: R2RPipelines,
        do_apply_cors: bool = True,
        *args,
        **kwargs,
    ):
        self.config = config
        self.providers = providers
        self.logging_connection = PipeLoggingConnectionSingleton()
        self.ingestion_pipeline = pipelines.ingestion_pipeline
        self.search_pipeline = pipelines.search_pipeline
        self.rag_pipeline = pipelines.rag_pipeline
        self.streaming_rag_pipeline = pipelines.streaming_rag_pipeline
        self.eval_pipeline = pipelines.eval_pipeline

        self.app = FastAPI()

        if do_apply_cors:
            R2RApp._apply_cors(self.app)

        self._setup_routes()

    def _setup_routes(self):
        self.app.add_api_route(
            path="/ingest_documents/",
            endpoint=self.ingest_documents_app,
            methods=["POST"],
        )
        self.app.add_api_route(
            path="/ingest_files/",
            endpoint=self.ingest_files_app,
            methods=["POST"],
        )
        self.app.add_api_route(
            path="/search/", endpoint=self.search_app, methods=["POST"]
        )
        self.app.add_api_route(
            path="/rag/", endpoint=self.rag_app, methods=["POST"]
        )
        self.app.add_api_route(
            path="/evaluate/",
            endpoint=self.evaluate_app,
            methods=["POST"],
        )
        self.app.add_api_route(
            path="/delete/", endpoint=self.delete_app, methods=["DELETE"]
        )
        self.app.add_api_route(
            path="/get_user_ids/",
            endpoint=self.get_user_ids_app,
            methods=["GET"],
        )
        self.app.add_api_route(
            path="/get_user_document_data/",
            endpoint=self.get_user_document_data_app,
            methods=["POST"],
        )
        self.app.add_api_route(
            path="/get_logs/",
            endpoint=self.get_logs_app,
            methods=["POST"],
        )

        self.app.add_api_route(
            path="/get_open_api_endpoint/",
            endpoint=self.get_open_api_endpoint,
            methods=["POST"],
        )
        # self.app.include_router(self.app.router)

    @syncable
    async def aingest_documents(self, documents: list[Document]):
        try:
            # Process the documents through the pipeline
            await self.ingestion_pipeline.run(
                input=to_async_generator(documents)
            )
            return {"results": "Entries upserted successfully."}
        except Exception as e:
            logger.error(
                f"ingest_documents(documents={documents}) - \n\n{str(e)})"
            )
            raise HTTPException(status_code=500, detail=str(e))

    class IngestDocumentsRequest(BaseModel):
        documents: list[Document]

    async def ingest_documents_app(self, request: IngestDocumentsRequest):
        try:
            return await self.aingest_documents(request.documents)
        except Exception as e:
            run_id = self.ingestion_pipeline.run_id or generate_run_id()
            await self.ingestion_pipeline.pipe_logger.log(
                pipe_run_id=run_id,
                key="pipeline_type",
                value=self.ingestion_pipeline.pipeline_type,
                is_pipeline_info=True,
            )

            await self.ingestion_pipeline.pipe_logger.log(
                pipe_run_id=run_id,
                key="error",
                value=str(e),
                is_pipeline_info=False,
            )
            raise HTTPException(status_code=500, detail=str(e))

    @syncable
    async def aingest_files(
        self,
        files: list[UploadFile],
        metadatas: Optional[list[dict]] = None,
        ids: Optional[list[uuid.UUID]] = None,
    ):
        if metadatas and len(metadatas) != len(files):
            raise HTTPException(
                status_code=400,
                detail="Number of metadata entries does not match number of files.",
            )
        if ids and len(ids) != len(files):
            raise HTTPException(
                status_code=400,
                detail="Number of ids does not match number of files.",
            )
        if len(files) == 0:
            raise HTTPException(
                status_code=400, detail="No files provided for ingestion."
            )

        try:
            documents = []
            for iteration, file in enumerate(files):
                logger.info(f"Processing file: {file.filename}")
                if (
                    file.size
                    > self.config.app.get("max_file_size_in_mb", 32)
                    * MB_CONVERSION_FACTOR
                ):
                    logger.error(f"File size exceeds limit: {file.filename}")
                    raise HTTPException(
                        status_code=413,
                        detail="File size exceeds maximum allowed size.",
                    )
                if not file.filename:
                    logger.error("File name not provided.")
                    raise HTTPException(
                        status_code=400, detail="File name not provided."
                    )

                file_content = await file.read()
                logger.info(f"File read successfully: {file.filename}")

                document_id = (
                    generate_id_from_label(file.filename)
                    if ids is None
                    else ids[iteration]
                )
                document_metadata = metadatas[iteration] if metadatas else {}

                documents.append(
                    Document(
                        id=document_id,
                        type=DocumentType(file.filename.split(".")[-1]),
                        data=file_content,
                        metadata=document_metadata,
                    )
                )
                logger.info(f"Document created: {document_id}")

            # Run the pipeline asynchronously
            logger.info("Running the ingestion pipeline...")
            await self.ingestion_pipeline.run(
                input=to_async_generator(documents),
            )
            logger.info("Ingestion pipeline completed.")

            return {
                "results": [
                    f"File '{file.filename}' processed successfully."
                    for file in files
                ]
            }
        except Exception as e:
            logger.error(
                f"ingest_files(metadata={metadatas}, ids={ids}, files={files}) - \n\n{str(e)})"
            )
            raise HTTPException(status_code=500, detail=str(e))
        finally:
            # Ensure all file handles are closed
            for file in files:
                file.file.close()

    async def ingest_files_app(
        self,
        files: list[UploadFile] = File(...),
        metadatas: Optional[str] = Form(None),
        ids: Optional[str] = Form(None),
    ):
        """Ingest files into the system."""
        try:
            if ids and ids != "null":
                ids_list = json.loads(ids)
                if len(ids_list) != 0:
                    try:
                        ids_list = [uuid.UUID(id) for id in ids_list]
                    except ValueError:
                        raise HTTPException(
                            status_code=400, detail="Invalid UUID provided."
                        )
            else:
                ids_list = None

            # Parse metadatas if provided
            metadatas = (
                json.loads(metadatas)
                if metadatas and metadatas != "null"
                else None
            )

            # Call aingest_files with the correct order of arguments
            return await self.aingest_files(
                files=files, metadatas=metadatas, ids=ids_list
            )
        except Exception as e:
            logger.error(f"ingest_files() - \n\n{str(e)})")
            run_id = self.ingestion_pipeline.run_id or generate_run_id()
            await self.ingestion_pipeline.pipe_logger.log(
                pipe_run_id=run_id,
                key="pipeline_type",
                value=self.ingestion_pipeline.pipeline_type,
                is_pipeline_info=True,
            )

            await self.ingestion_pipeline.pipe_logger.log(
                pipe_run_id=run_id,
                key="error",
                value=str(e),
                is_pipeline_info=False,
            )
            raise HTTPException(status_code=500, detail=str(e))

    @syncable
    async def asearch(
        self,
        query: str,
        search_filters: Optional[dict] = None,
        search_limit: int = 10,
    ):
        """Search for documents based on the query."""
        try:
            search_filters = search_filters or {}
            results = await self.search_pipeline.run(
                input=to_async_generator([query]),
                search_filters=search_filters,
                search_limit=search_limit,
            )
            return {"results": [results.dict() for results in results]}
        except Exception as e:
            logger.error(f"search(query={query}) - \n\n{str(e)})")
            raise HTTPException(status_code=500, detail=str(e))

    class SearchRequest(BaseModel):
        query: str
        search_filters: Optional[str] = None
        search_limit: int = 10

    async def search_app(self, request: SearchRequest):
        try:
            search_filters = (
                {}
                if request.search_filters is None
                or request.search_filters == "null"
                else json.loads(request.search_filters)
            )
            return await self.asearch(
                request.query, search_filters, request.search_limit
            )
        except Exception as e:
            # TODO - Make this more modular
            run_id = self.search_pipeline.run_id or generate_run_id()
            await self.search_pipeline.pipe_logger.log(
                pipe_run_id=run_id,
                key="pipeline_type",
                value=self.search_pipeline.pipeline_type,
                is_pipeline_info=True,
            )

            await self.search_pipeline.pipe_logger.log(
                pipe_run_id=run_id,
                key="error",
                value=str(e),
                is_pipeline_info=False,
            )
            raise HTTPException(status_code=500, detail=str(e))

    @syncable
    async def arag(
        self,
        message: str,
        rag_generation_config: GenerationConfig,
        search_filters: Optional[dict[str, str]] = None,
        search_limit: int = 10,
    ):
        try:
            if rag_generation_config.stream:

                async def stream_response():
                    async for chunk in await self.streaming_rag_pipeline.run(
                        input=to_async_generator([message]),
                        streaming=True,
                        search_filters=search_filters,
                        search_limit=search_limit,
                        rag_generation_config=rag_generation_config,
                    ):
                        yield chunk

                return stream_response()

            else:
                results = await self.rag_pipeline.run(
                    input=to_async_generator([message]),
                    streaming=False,
                    search_filters=search_filters,
                    search_limit=search_limit,
                    rag_generation_config=rag_generation_config,
                )
                return results
        except Exception as e:
            logger.error(f"rag(message={message}) - \n\n{str(e)})")
            raise HTTPException(status_code=500, detail=str(e))

    class RAGRequest(BaseModel):
        message: str
        search_filters: Optional[str] = None
        search_limit: int = 10
        rag_generation_config: Optional[str] = None
        streaming: bool = False

    async def rag_app(self, request: RAGRequest):
        try:
            search_filters = (
                None
                if request.search_filters is None
                or request.search_filters == "null"
                else json.loads(request.search_filters)
            )
            rag_generation_config = (
                GenerationConfig(
                    **json.loads(request.rag_generation_config),
                    stream=request.streaming,
                )
                if request.rag_generation_config
                and request.rag_generation_config != "null"
                else GenerationConfig(
                    model="gpt-3.5-turbo", stream=request.streaming
                )
            )
            response = await self.arag(
                request.message,
                rag_generation_config,
                search_filters,
                request.search_limit,
            )
            if request.streaming:
                return StreamingResponse(
                    response, media_type="application/json"
                )
            else:
                return {"results": response}

        except Exception as e:
            # TODO - Make this more modular
            run_id = self.rag_pipeline.run_id or generate_run_id()
            await self.rag_pipeline.pipe_logger.log(
                pipe_run_id=run_id,
                key="pipeline_type",
                value=self.rag_pipeline.pipeline_type,
                is_pipeline_info=True,
            )

            await self.rag_pipeline.pipe_logger.log(
                pipe_run_id=run_id,
                key="error",
                value=str(e),
                is_pipeline_info=False,
            )
            raise HTTPException(status_code=500, detail=str(e))

    @syncable
    async def aevaluate(
        self,
        query: str,
        context: str,
        completion: str,
    ):
        try:
            eval_payload = R2REvalPipe.EvalPayload(
                query=query,
                context=context,
                completion=completion,
            )
            result = await self.eval_pipeline.run(
                input=to_async_generator([eval_payload])
            )
            return {"results": result}
        except Exception as e:
            logger.error(f"evaluate(query={query}) - \n\n{str(e)})")
            raise HTTPException(status_code=500, detail=str(e))

    class EvalRequest(BaseModel):
        query: str
        context: str
        completion: str

    async def evaluate_app(self, request: EvalRequest):
        try:
            return await self.aevaluate(
                query=request.query,
                context=request.context,
                completion=request.completion,
            )
        except Exception as e:
            run_id = self.eval_pipeline.run_id or generate_run_id()
            await self.eval_pipeline.pipe_logger.log(
                pipe_run_id=run_id,
                key="pipeline_type",
                value=self.eval_pipeline.pipeline_type,
                is_pipeline_info=True,
            )

            await self.eval_pipeline.pipe_logger.log(
                pipe_run_id=run_id,
                key="error",
                value=str(e),
                is_pipeline_info=False,
            )
            raise HTTPException(status_code=500, detail=str(e))

    @syncable
    async def adelete(self, key: str, value: Union[bool, int, str]):
        try:
            self.providers.vector_db.delete_by_metadata(key, value)
            return {"results": "Entries deleted successfully."}
        except Exception as e:
            logger.error(
                f":delete: [Error](key={key}, value={value}, error={str(e)})"
            )
            raise HTTPException(status_code=500, detail=str(e))

    class DeleteRequest(BaseModel):
        key: str
        value: Union[bool, int, str]

    async def delete_app(self, request: DeleteRequest = Body(...)):
        return await self.adelete(request.key, request.value)

    @syncable
    async def aget_user_ids(self):
        try:
            user_ids = self.providers.vector_db.get_metadatas(
                metadata_fields=["user_id"]
            )

            return {"results": [ele["user_id"] for ele in user_ids]}
        except Exception as e:
            logger.error(f"get_user_ids() - \n\n{str(e)})")
            raise HTTPException(status_code=500, detail=str(e))

    async def get_user_ids_app(self):
        return await self.aget_user_ids()

    @syncable
    async def aget_user_document_data(self, user_id: str):
        try:
            if isinstance(user_id, uuid.UUID):
                user_id = str(user_id)
            document_ids = self.providers.vector_db.get_metadatas(
                metadata_fields=["document_id", "title"],
                filter_field="user_id",
                filter_value=user_id,
            )
            return {"results": [ele for ele in document_ids]}
        except Exception as e:
            logger.error(
                f"get_user_document_data(user_id={user_id}) - \n\n{str(e)})"
            )
            raise HTTPException(status_code=500, detail=str(e))

    class UserDocumentRequest(BaseModel):
        user_id: str

    async def get_user_document_data_app(self, request: UserDocumentRequest):
        return await self.aget_user_document_data(request.user_id)

    @syncable
    async def aget_logs(self, pipeline_type: Optional[str] = None):
        try:
            logs_per_run = 10
            if self.logging_connection is None:
                raise HTTPException(
                    status_code=404, detail="Logging provider not found."
                )
            run_info = await self.logging_connection.get_run_info(
                pipeline_type=pipeline_type,
                limit=self.config.app.get("max_logs", 100) // logs_per_run,
            )
            run_ids = [run.run_id for run in run_info]
            if len(run_ids) == 0:
                return {"results": []}
            logs = await self.logging_connection.get_logs(run_ids)
            # Aggregate logs by run_id and include run_type
            aggregated_logs = []

            for run in run_info:
                run_logs = [
                    log for log in logs if log["pipe_run_id"] == run.run_id
                ]
                entries = [
                    {"key": log["key"], "value": log["value"]}
                    for log in run_logs
                ]
                aggregated_logs.append(
                    {
                        "run_id": run.run_id,
                        "run_type": run.pipeline_type,
                        "entries": entries,
                    }
                )

            return {"results": aggregated_logs}

        except Exception as e:
            logger.error(f":logs: [Error](error={str(e)})")
            raise HTTPException(status_code=500, detail=str(e))

    class LogsRequest(BaseModel):
        pipeline_type: Optional[str] = None

    async def get_logs_app(self, request: LogsRequest):
        return await self.aget_logs(request.pipeline_type)

    def get_open_api_endpoint(self):
        from fastapi.openapi.utils import get_openapi

        return {
            "results": get_openapi(
                title="R2R Application API",
                version="1.0.0",
                routes=self.app.routes,
            )
        }

    def serve(self, host: str = "0.0.0.0", port: int = 8000):
        try:
            import uvicorn
        except ImportError:
            raise ImportError(
                "Please install uvicorn using 'pip install uvicorn'"
            )

        uvicorn.run(self.app, host=host, port=port)

    @staticmethod
    def _apply_cors(app):
        # CORS setup
        origins = [
            "*",  # TODO - Change this to the actual frontend URL
            "http://localhost:3000",
            "http://localhost:8000",
        ]

        app.add_middleware(
            CORSMiddleware,
            allow_origins=origins,  # Allows specified origins
            allow_credentials=True,
            allow_methods=["*"],  # Allows all methods
            allow_headers=["*"],  # Allows all headers
        )
